import requests
import json
import abc
import logging

from requests.auth import HTTPBasicAuth


class RexCoreValueReader:

    STATE_UNCONFIGURED = 'unconfigured'
    STATE_CONFIGURED = 'configured'
    STATE_LOADED = 'loaded'

    def __init__(self):
        self.connections = {}
        self.payload = {}
        self.response = {}

    def read(self, ids):
        if ids:
            self._init_response(ids)

        for connection_id in self.connections:
            self._read_rexcore_items(connection_id)

    def result(self):
        return self.response

    def _init_response(self, ids):
        signals_ids = ids.split(',')

        for signal_id in signals_ids:
            signal_id = int(signal_id)

            self.response[signal_id] = {
                'value': None,
                'state': self.STATE_UNCONFIGURED,
            }

            self._append_rexcore_item(signal_id)

    def _append_rexcore_item(self, signal_id):
        signal = self._get_signal_object(signal_id)
        if signal is None:
            return

        if not signal.rex_value_path:
            return

        if signal.rexcore_id not in self.connections:
            rexcore = self._get_connection_object(signal.rexcore_id)
            if rexcore is not None:
                self.connections[rexcore.id] = rexcore

        if signal.rexcore_id in self.connections:
            rexcore = self.connections[signal.rexcore_id]

            if rexcore.id not in self.payload:
                self.payload[rexcore.id] = []

            self.response[signal_id]['state'] = self.STATE_CONFIGURED
            self.payload[rexcore.id].append(signal)

    def _read_rexcore_items(self, connection_id):
        rexcore = self.connections[connection_id]

        if not rexcore.connection.startswith("http"):
            rexcore.connection = "http://" + rexcore.connection

        if not rexcore.connection.endswith("api/rpc"):
            rexcore.connection = rexcore.connection + "/api/rpc"

        names = []
        signals = self.payload[connection_id]

        for signal in signals:
            rex_value_path = signal.rex_value_path
            names.append(rex_value_path)

        api_response = self._read_rexcore_values(rexcore, names)
        self._parse_rexcore_response(connection_id, api_response)

    def _read_rexcore_values(self, rexcore, names):
        payload = {
            "method": "ReadItems",
            "params": {
                "format": "long",
                "names": names,
            },
        }

        try:
            api_response = requests.post(
                rexcore.connection,
                timeout=10,
                data=json.dumps(payload),
                headers={'content-type': 'application/json'},
                auth=HTTPBasicAuth(rexcore.username, rexcore.password)
            ).json()

            return api_response
        except Exception as e:
            logging.error(e)
            return None

    def _parse_rexcore_response(self, connection_id, api_response):
        if api_response and isinstance(api_response, dict) and 'result' in api_response:
            result = api_response['result']

            if isinstance(result, dict) and 'values' in result:
                values = result['values']
                index = 0

                signals = self.payload[connection_id]
                for signal in signals:
                    try:
                        v = values[index]['v']
                        if type(v) == int or type(v) == float:
                            signal_id = signal.id
                            self.response[signal_id]['value'] = v
                            self.response[signal_id]['state'] = self.STATE_LOADED
                    except IndexError:
                        pass
                    except KeyError:
                        pass

                    index = index + 1

    @abc.abstractmethod
    def _get_signal_object(self, signal_id):
        pass

    @abc.abstractmethod
    def _get_connection_object(self, rexcore_id):
        pass
