from copy import copy
from enum import Enum

from ..core.data_provider import DataProvider

import datetime


class AggregationInterval(Enum):
    V_NONE = "0"
    V_5_MINUTES = "5"
    V_15_MINUTES = "15"
    V_1_HOUR = "60"
    V_4_HOURS = "240"
    V_12_HOURS = "720"
    V_24_HOURS = "1440"
    V_7_DAYS = "10080"
    V_15_DAYS = "21600"
    V_30_DAYS = "43200"

    def __str__(self):
        return str(self.value)


class ExportType(Enum):
    SHEET_BY_TYPE = "by_type"
    SHEET_PER_SIGNAL = "per_signal"

    def __str__(self):
        return self.value


class Export:
    def __init__(self, provider):
        self.provider = provider

    def write_header(self, sheet, headers, formats=[], row_index=1):
        index = 0

        for cell in headers:
            index = index + 1

            cell = sheet.cell(column=index, row=row_index, value=cell)

            font = copy(cell.font)
            font.bold = True
            cell.font = font

        sheet.auto_filter.ref = sheet.dimensions

        for column_cells in sheet.columns:
            column = column_cells[0]

            sheet.column_dimensions[column.column_letter].width = self.get_column_width()

    def write_format(self, sheet, formats=[]):
        index = -1

        for column_cells in sheet.columns:
            column = column_cells[0]
            index = index + 1

            for cell in sheet[column.column_letter]:
                cell.number_format = formats[index]

    def get_column_width(self):
        return 15

    def date_format(self):
        return 'YYYY-MM-DD'

    def time_format(self):
        return 'hh:mm:ss.000'

    def create_number_format(self, signal):
        format = "0" * signal.decimals
        return "0." + format

    def quality_format(self):
        return '0.00'

    def count_format(self):
        return '0'

    def generate_values(self, context, signal):
        data = self.provider.load(context, signal)
        for item in data:
            yield item

    def generate_values_signals(self, context, signals):  # TODO MM
        data = self.provider.load_multi(context, signals)
        for item in data:
            yield item


class ExportNoAggregation(Export):
    def get_date_time_header(self):
        return [
            "date",
            "time",
        ]

    def get_date_time_header_format(self):
        return [
            self.date_format(),
            self.time_format(),
        ]


class ExportWithAggregation(Export):
    def get_date_time_header(self):
        return [
            "date_begin",
            "time_begin",
            "date_end",
            "time_end",
        ]

    def get_date_time_header_format(self):
        return [
            self.date_format(),
            self.time_format(),
            self.date_format(),
            self.time_format(),
        ]

    def window_fn(self, time, dt_start, window_segment):
        diff = time - dt_start
        diff_seconds = diff.total_seconds()
        return int(diff_seconds / window_segment)

    def generate_date_time_columns(self, sheet, context):
        dt_start = context['dt_start']
        dt_end = context['dt_end']
        window_segment = int(context['aggregation_interval'].value) * 60
        index = 0

        while True:
            stamp = dt_start.timestamp() + window_segment * index
            if stamp >= dt_end.timestamp():
                break

            seg_dt_start = datetime.datetime.fromtimestamp(stamp)
            seg_dt_end = datetime.datetime.fromtimestamp(stamp + window_segment)

            window_index = self.window_fn(seg_dt_start, dt_start, window_segment)
            row = self.get_row_index(window_index)

            sheet.cell(column=1, row=row, value=seg_dt_start)
            sheet.cell(column=2, row=row, value=seg_dt_start)
            sheet.cell(column=3, row=row, value=seg_dt_end)
            sheet.cell(column=4, row=row, value=seg_dt_end)

            index = index + 1

        return dt_start, window_segment

    def get_row_index(self, index):
        return index + 2
