import datetime

from django.test import TestCase

from ...core.aggregator import Aggregator


class AggregatorTestCase(TestCase):

    def test_aggregator_leave_original_data_when_no_aggregation_processed(self):
        context = {
            'dt_start': datetime.datetime(2021, 1, 1, 0, 0),
            'dt_end': datetime.datetime(2021, 1, 31, 0, 0),
            'level': 31,
        }

        context = self.prepare_context(context)
        data = [[
            datetime.datetime(2021, 1, 1, 12, 0),
            100.00,
            50.00,
            150.00,
            1.00,
        ], [
            datetime.datetime(2021, 1, 2, 12, 0),
            150.00,
            100.00,
            200.00,
            1.00,
        ]]

        aggregator = Aggregator()
        aggregator.agregate(context, data)

        window_level, values = aggregator.result()

        self.assertEqual(window_level, 0)
        self.assertEqual(values, [{
            'cnt': 1,
            'dt': datetime.datetime(2021, 1, 1, 12, 0),
            'level': 0,
            'quality': 1.0,
            'value': 100.0,
            'value_max': 150.0,
            'value_min': 50.0,
        }, {
            'cnt': 1,
            'dt': datetime.datetime(2021, 1, 2, 12, 0),
            'level': 1,
            'quality': 1.0,
            'value': 150.0,
            'value_max': 200.0,
            'value_min': 100.0,
        }])

    def test_aggregator_leave_original_data_in_small_window(self):
        context = {
            'dt_start': datetime.datetime(2021, 1, 12, 13, 43, 13),
            'dt_end': datetime.datetime(2021, 1, 12, 13, 51, 54),
            'level': 1920,
        }

        context = self.prepare_context(context)
        data = [[
            datetime.datetime(2021, 1, 12, 13, 45, 0),
            100.00,
            50.00,
            150.00,
            1.00,
        ], [
            datetime.datetime(2021, 1, 12, 13, 50, 0),
            150.00,
            100.00,
            200.00,
            1.00,
        ]]

        aggregator = Aggregator()
        aggregator.agregate(context, data)

        window_level, values = aggregator.result()

        self.maxDiff = None
        self.assertEqual(window_level, 0)
        self.assertEqual(values, [{
            'cnt': 1,
            'dt': datetime.datetime(2021, 1, 12, 13, 45, 0),
            'level': 107,
            'quality': 1.0,
            'value': 100.0,
            'value_max': 150.0,
            'value_min': 50.0,
        }, {
            'cnt': 1,
            'dt': datetime.datetime(2021, 1, 12, 13, 50, 0),
            'level': 407,
            'quality': 1.0,
            'value': 150.0,
            'value_max': 200.0,
            'value_min': 100.0,
        }])

    def test_aggregator_aggregated_values_and_stamps_moved_into_window(self):
        context = {
            'dt_start': datetime.datetime(2021, 1, 1, 0, 0),
            'dt_end': datetime.datetime(2021, 1, 31, 0, 0),
            'level': 7,
        }

        context = self.prepare_context(context)
        data = [[
            datetime.datetime(2021, 1, 1, 12, 0),
            100.00,
            50.00,
            150.00,
            1.00,
        ], [
            datetime.datetime(2021, 1, 2, 12, 0),
            150.00,
            100.00,
            200.00,
            1.00,
        ]]

        aggregator = Aggregator()
        aggregator.agregate(context, data)

        window_level, values = aggregator.result()

        self.assertEqual(window_level, 1)
        self.assertEqual(values, [{
            'cnt': 2,
            'dt': datetime.datetime(2021, 1, 1, 0, 0),
            'level': 0,
            'quality': 1.0,
            'value': 125.0,
            'value_max': 200.0,
            'value_min': 50.0,
        }])

    def prepare_context(self, context):
        dt_start = context['dt_start']
        dt_end = context['dt_end']
        level = context['level']

        dt_diff = dt_end - dt_start
        dt_diff_seconds = dt_diff.total_seconds()

        window_segment = int(dt_diff_seconds / level) + 1 # TODO MM move into helper function
        context['window_segment'] = window_segment

        return context
