import abc
import datetime
import json

from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.serializers import ValidationError

from django.conf import settings
from django.http import HttpResponse

from openpyxl import Workbook
from openpyxl.writer.excel import save_virtual_workbook

from .serializers import SignalSerializer, GroupSerializer, SignalGroupSerializer, \
        RexCoreSerializer
from .authentication import IsAuthenticated

from .export.core import AggregationInterval, ExportType
from .export.signal_analog_no_aggregation import SignalAnalogNoAggregation
from .export.signal_cumulative_no_aggregation import SignalCumulativeNoAggregation
from .export.signal_bool_no_aggregation import SignalBoolNoAggregation
from .export.signal_analog_with_aggregation import SignalAnalogWithAggregation
from .export.signal_cumulative_with_aggregation import SignalCumulativeWithAggregation
from .export.signal_bool_with_aggregation import SignalBoolWithAggregation
from .export.signals_analog_no_aggregation import SignalsAnalogNoAggregation
from .export.signals_cumulative_no_aggregation import SignalsCumulativeNoAggregation
from .export.signals_bool_no_aggregation import SignalsBoolNoAggregation
from .export.signals_analog_with_aggregation import SignalsAnalogWithAggregation
from .export.signals_cumulative_with_aggregation import SignalsCumulativeWithAggregation
from .export.signals_bool_with_aggregation import SignalsBoolWithAggregation
from .export.additional_sheets import NodeSheet, SignalSheet


class Option(APIView):
    permission_classes = [IsAuthenticated]

    def __init__(self):
        self.option = self.define_option()

    def define_option(self):
        if hasattr(settings, 'APP_OPTION'):
            return settings.APP_OPTION
        else:
            return {}

    def get(self, request):
        return Response(self.option)


class Version(APIView):
    permission_classes = [IsAuthenticated]

    def get(self, request):
        if not hasattr(settings, 'STATIC_VERSION_FILE'):
            raise Exception("Missing version file configuration!")

        try:
            file = settings.STATIC_VERSION_FILE
            with open(file) as version_file:
                version = json.load(version_file)
        except IOError:
            raise Exception("Missing version file!")
        except Exception:
            raise Exception("Invalid version file!")

        return Response(version)


class SignalList(APIView):
    permission_classes = [IsAuthenticated]

    @abc.abstractmethod
    def get_signals(self):
        pass

    def get(self, request):
        signals = self.get_signals()
        serializer = SignalSerializer(signals, many=True)
        return Response(serializer.data)


class SignalDetail(APIView):
    permission_classes = [IsAuthenticated]

    @abc.abstractmethod
    def get_signal(self, pk):
        pass

    def get(self, request, signal_pk):
        signal = self.get_signal(signal_pk)
        serializer = SignalSerializer(signal)
        return Response(serializer.data)


class GroupList(APIView):
    permission_classes = [IsAuthenticated]

    @abc.abstractmethod
    def get_groups(self):
        pass

    def get(self, request):
        groups = self.get_groups()
        serializer = GroupSerializer(groups, many=True)
        return Response(serializer.data)


class GroupDetail(APIView):
    permission_classes = [IsAuthenticated]

    @abc.abstractmethod
    def get_group(self, pk):
        pass

    def get(self, request, group_pk):
        group = self.get_group(group_pk)
        serializer = GroupSerializer(group)
        return Response(serializer.data)


class SignalGroupsList(APIView):
    permission_classes = [IsAuthenticated]

    @abc.abstractmethod
    def get_signal_groups(self, signal_pk):
        pass

    def get(self, request, signal_pk):
        groups = self.get_signal_groups(signal_pk)
        serializer = SignalGroupSerializer(groups, many=True)
        return Response(serializer.data)


class RexCoreList(APIView):
    permission_classes = [IsAuthenticated]

    @abc.abstractmethod
    def get_connections(self):
        pass

    def get(self, request):
        groups = self.get_connections()
        serializer = RexCoreSerializer(groups, many=True)
        return Response(serializer.data)


class RexCoreDetail(APIView):
    permission_classes = [IsAuthenticated]

    @abc.abstractmethod
    def get_connection(self, pk):
        pass

    def get(self, request, group_pk):
        group = self.get_connection(group_pk)
        serializer = RexCoreSerializer(group)
        return Response(serializer.data)


class BatchSignalLastValue(APIView):
    permission_classes = [IsAuthenticated]

    @abc.abstractmethod
    def _create_rexcore_value_reader(self):
        pass

    def get(self, request):
        ids = self.request.query_params.get('ids', None)

        reader = self._create_rexcore_value_reader()
        reader.read(ids)

        response = reader.result()

        return Response(response)


class SignalDataView(APIView):
    permission_classes = [IsAuthenticated]

    def get_context(self):
        dt_start = self.request.query_params.get('dt_start', None)
        if dt_start is None:
            raise ValidationError('Field dt_start is required!')

        dt_start = datetime.datetime.strptime(dt_start, '%Y-%m-%d-%H-%M-%S') # TODO MM format
        if dt_start is None:
            raise ValidationError('dt_start has incorect format!')

        dt_end = self.request.query_params.get('dt_end', None)
        if dt_end is None:
            raise ValidationError('Field dt_end is required!')

        dt_end = datetime.datetime.strptime(dt_end, '%Y-%m-%d-%H-%M-%S') # TODO MM format
        if dt_end is None:
            raise ValidationError('dt_end has incorect format!')

        dt_diff = dt_end - dt_start
        dt_diff_seconds = dt_diff.total_seconds() # TODO MM move into helper function
        if dt_diff_seconds <= 0:
            raise ValidationError('Date range is not valid!')

        level = self.request.query_params.get('level', None)
        if level is None:
            raise ValidationError('Field level is required!')
        if not level.isnumeric():
            raise ValidationError('Level must be number!')
        level = int(level)
        if level <= 0:
            raise ValidationError('Level must be positive number!')

        window_segment = int(dt_diff_seconds / level) + 1 # TODO MM move into helper function
        if window_segment <= 0:
            raise ValidationError('Date range is not valid!')

        return {
            'dt_start': dt_start,
            'dt_end': dt_end,
            'level': level,
            'window_segment': window_segment,
        }

    @abc.abstractmethod
    def get(self, request, signal_pk):
        pass


class SignalDataAutoView(SignalDataView):
    pass


class SignalDataAnalogView(SignalDataView):
    pass


class SignalDataBarView(SignalDataView):
    pass


class SignalDataCumulativeView(SignalDataView):
    pass


class SignalDataBoolView(SignalDataView):
    pass


class Export(APIView):
    permission_classes = [IsAuthenticated]

    def get_context(self):
        dt_start = self.request.query_params.get('dt_start', None)
        if dt_start is None:
            raise ValidationError('Field dt_start is required!')

        dt_start = datetime.datetime.strptime(dt_start, '%Y-%m-%d-%H-%M-%S')  # TODO MM format
        if dt_start is None:
            raise ValidationError('dt_start has incorect format!')

        dt_end = self.request.query_params.get('dt_end', None)
        if dt_end is None:
            raise ValidationError('Field dt_end is required!')

        dt_end = datetime.datetime.strptime(dt_end, '%Y-%m-%d-%H-%M-%S')  # TODO MM format
        if dt_end is None:
            raise ValidationError('dt_end has incorect format!')

        dt_diff = dt_end - dt_start
        dt_diff_seconds = dt_diff.total_seconds()  # TODO MM move into helper function
        if dt_diff_seconds <= 0:
            raise ValidationError('Date range is not valid!')

        aggregation = self.request.query_params.get('aggregation', None)
        if aggregation is None:
            raise ValidationError('Field aggregation is required!')
        if not aggregation.isnumeric():
            raise ValidationError('Aggregation must be number!')
        aggregation = int(aggregation)
        if aggregation < 0:
            raise ValidationError('Aggregation must be positive number!')
        aggregation = AggregationInterval(str(aggregation))

        signals_ids = self.request.query_params.get('signals_ids', None)
        if signals_ids is None:
            raise ValidationError('Field signals_ids is required!')

        export_type = self.request.query_params.get('export_type', None)
        if export_type is None:
            raise ValidationError('Field export_type is required!')

        helper_node = self.request.query_params.get('helper_node', None)
        helper_signal = self.request.query_params.get('helper_signal', None)

        if not signals_ids and not helper_node and not helper_signal:
            raise ValidationError('At least one field of signals_ids or helper_sheets is required!')

        return {
            'dt_start': dt_start,
            'dt_end': dt_end,
            'aggregation_interval': aggregation,
            'signals_ids': signals_ids,
            'export_type': export_type,
            'helper_node': helper_node,
            'helper_signal': helper_signal,
        }

    @abc.abstractmethod
    def get_nodes(self):
        pass

    @abc.abstractmethod
    def get_signals(self):
        pass

    @abc.abstractmethod
    def get_signal(self, pk):
        pass

    @abc.abstractmethod
    def get_data_provider(self):
        pass

    def get(self, request):
        try:
            context = self.get_context()

            signals_ids = context['signals_ids'].split(",")
            aggregation_interval = context['aggregation_interval']
            export_type = context['export_type']

            types = {}
            for signal_id in signals_ids:
                signal = self.get_signal(int(signal_id))
                if signal.type not in types:
                    types[signal.type] = []

                types[signal.type].append(signal)

            wb = Workbook()
            wb.remove_sheet(wb.active)

            if len(signals_ids) > 0:
                if aggregation_interval == AggregationInterval.V_NONE:
                    if export_type == ExportType.SHEET_BY_TYPE.value:
                        for type in types:
                            signals = types[type]

                            if type == "ANALOG":
                                sheetGenerator = SignalsAnalogNoAggregation(self.get_data_provider())
                                sheet = wb.create_sheet(title='DATA_ANALOG')
                                sheetGenerator.create_sheet(wb, sheet, context, signals)

                            elif type == "CUMULATIVE":
                                sheetGenerator = SignalsCumulativeNoAggregation(self.get_data_provider())
                                sheet = wb.create_sheet(title='DATA_CUMULATIVE')
                                sheetGenerator.create_sheet(wb, sheet, context, signals)

                            elif type == "BOOL":
                                sheetGenerator = SignalsBoolNoAggregation(self.get_data_provider())
                                sheet = wb.create_sheet(title='DATA_BOOL')
                                sheetGenerator.create_sheet(wb, sheet, context, signals)

                    if export_type == ExportType.SHEET_PER_SIGNAL.value:
                        for type in types:
                            for signal in types[type]:
                                sheet_name = signal.name
                                sheet_name = sheet_name.replace("/", "_")
                                sheet_name = sheet_name.replace(":", "_")
                                sheet_name = sheet_name.replace("[", "_")
                                sheet_name = sheet_name.replace("]", "_")
                                sheet = wb.create_sheet(title=sheet_name)

                                if type == "ANALOG":
                                    sheetGenerator = SignalAnalogNoAggregation(self.get_data_provider())
                                    sheetGenerator.create_sheet(wb, sheet, context, signal)

                                elif type == "CUMULATIVE":
                                    sheetGenerator = SignalCumulativeNoAggregation(self.get_data_provider())
                                    sheetGenerator.create_sheet(wb, sheet, context, signal)

                                elif type == "BOOL":
                                    sheetGenerator = SignalBoolNoAggregation(self.get_data_provider())
                                    sheetGenerator.create_sheet(wb, sheet, context, signal)
                else:
                    if export_type == ExportType.SHEET_BY_TYPE.value:
                        for type in types:
                            signals = types[type]

                            if type == "ANALOG":
                                sheetGenerator = SignalsAnalogWithAggregation(self.get_data_provider())
                                sheet = wb.create_sheet(title='DATA_ANALOG')
                                sheetGenerator.create_sheet(wb, sheet, context, signals)

                            elif type == "CUMULATIVE":
                                sheetGenerator = SignalsCumulativeWithAggregation(self.get_data_provider())
                                sheet = wb.create_sheet(title='DATA_CUMULATIVE')
                                sheetGenerator.create_sheet(wb, sheet, context, signals)

                            elif type == "BOOL":
                                sheetGenerator = SignalsBoolWithAggregation(self.get_data_provider())
                                sheet = wb.create_sheet(title='DATA_BOOL')
                                sheetGenerator.create_sheet(wb, sheet, context, signals)

                    if export_type == ExportType.SHEET_PER_SIGNAL.value:
                        for type in types:
                            for signal in types[type]:
                                sheet_name = signal.name
                                sheet_name = sheet_name.replace("/", "_")
                                sheet_name = sheet_name.replace(":", "_")
                                sheet_name = sheet_name.replace("[", "_")
                                sheet_name = sheet_name.replace("]", "_")
                                sheet = wb.create_sheet(title=sheet_name)

                                if type == "ANALOG":
                                    sheetGenerator = SignalAnalogWithAggregation(self.get_data_provider())
                                    sheetGenerator.create_sheet(wb, sheet, context, signal)

                                elif type == "CUMULATIVE":
                                    sheetGenerator = SignalCumulativeWithAggregation(self.get_data_provider())
                                    sheetGenerator.create_sheet(wb, sheet, context, signal)

                                elif type == "BOOL":
                                    sheetGenerator = SignalBoolWithAggregation(self.get_data_provider())
                                    sheetGenerator.create_sheet(wb, sheet, context, signal)

            if context['helper_node']:
                sheetGenerator = NodeSheet(self.get_data_provider(), self.get_nodes())
                sheetGenerator.create_sheet(wb)

            if context["helper_signal"]:
                sheetGenerator = SignalSheet(self.get_data_provider(), self.get_signals())
                sheetGenerator.create_sheet(wb)

            response = HttpResponse(
                content=save_virtual_workbook(wb),
                content_type='application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
            )
            response['Content-Disposition'] = 'attachment; filename=export.xlsx'
            return response

        except ValidationError as e:
            # TODO MM
            raise e
