from django.db import models

from datetime import datetime
from bulk_sync import bulk_sync


class Data(models.Model):
    id = models.BigIntegerField(primary_key=True)
    signal = models.ForeignKey('Signal', models.DO_NOTHING)
    arcid = models.BigIntegerField()
    dt = models.DateTimeField()
    value = models.FloatField()
    value_min = models.FloatField(blank=True, null=True)
    value_max = models.FloatField(blank=True, null=True)
    quality = models.FloatField()
    dt_prev = models.FloatField(blank=True, null=True)
    dt_next = models.FloatField(blank=True, null=True)

    class Meta:
        abstract = True
        managed = True


def get_data_model(base, key):
    class_name = "DataModel_%s" % key

    DataModel = type(
        class_name,
        (base,),
        {
            '__module__': __name__,
            'Meta': type(
                "Meta",
                (),
                {
                    'db_table': key
                }
            ),
        }
    )

    return DataModel


def create_data_models(parent_class, stamp_first, stamp_last):
    stamp_first_o = datetime.strptime(stamp_first + "01", '%Y%m%d')
    stamp_last_o = datetime.strptime(stamp_last + "01", '%Y%m%d')

    stamp_actual = stamp_first_o

    while stamp_actual <= stamp_last_o:
        model_name = "DATA_TS1_" + stamp_actual.strftime("%Y%m")
        get_data_model(parent_class, model_name)

        stamp_actual_year = int(stamp_actual.strftime("%Y"))
        stamp_actual_month = int(stamp_actual.strftime("%m"))

        stamp_actual_month = stamp_actual_month + 1
        if stamp_actual_month > 12:
            stamp_actual_year = stamp_actual_year + 1
            stamp_actual_month = 1

        stamp_actual = datetime.strptime(
            str(stamp_actual_year).zfill(2) + \
            str(stamp_actual_month).zfill(2) + \
            "01",
            '%Y%m%d'
        )


class Datatbl(models.Model):
    id = models.FloatField(primary_key=True)
    tblspace = models.ForeignKey('Tblspace', models.DO_NOTHING)
    date_from = models.DateField()
    date_to = models.DateField(blank=True, null=True)
    name = models.CharField(max_length=32)

    class Meta:
        managed = True
        db_table = 'datatbl'


class Grp(models.Model):
    id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=128)
    description = models.CharField(max_length=1024)

    class Meta:
        managed = True
        db_table = 'grp'

    def get_all_groups():
        groups = Grp.objects.all()
        return groups

    def get_one_group_by_pk(pk):
        return Grp.objects.get(pk=pk)

    def get_signals_groups(signal_pk):
        return Grp.objects.filter(signalgrp__signal=signal_pk)


class Node(models.Model):
    id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=128)
    description = models.CharField(max_length=1024, blank=True, null=True)
    connection = models.CharField(max_length=1024, blank=True, null=True)

    class Meta:
        managed = True
        db_table = 'node'

    def get_all_nodes():
        nodes = Node.objects.all()
        return nodes


class Rexcore(models.Model):
    id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=128)
    description = models.CharField(max_length=1024, blank=True, null=True)
    connection = models.CharField(max_length=1024)
    username = models.CharField(max_length=1024)
    password = models.CharField(max_length=1024, blank=True, null=True)

    class Meta:
        managed = True
        db_table = 'rexcore'

    def get_all_connections():
        connections = Rexcore.objects.all()
        return connections

    def get_one_connection_by_pk(pk):
        return Rexcore.objects.get(pk=pk)

    def update_all_connections(connections):
        # all objects
        objects_all = []

        for connection in connections:
            object = Rexcore(**connection)
            objects_all.append(object)

        bulk_sync(
            new_models=objects_all if len(objects_all) > 0 else Rexcore.objects.none(),
            filters=[],
            fields=[
                'name',
                'description',
                'connection',
                'username',
            ],
            key_fields=['id']
        )

        # change password
        objects_password = []

        for connection in connections:
            object = Rexcore(**connection)
            if object.password is not None:
                objects_password.append(object)

        if len(objects_password) > 0:
            Rexcore.objects.bulk_update(
                objects_password,
                ['password']
            )


class Signal(models.Model):
    id = models.BigIntegerField(primary_key=True)
    node = models.ForeignKey(Node, models.DO_NOTHING)
    tblspace = models.ForeignKey('Tblspace', models.DO_NOTHING)
    name = models.CharField(max_length=128)
    description = models.CharField(max_length=1024)
    type = models.CharField(max_length=32)
    unit = models.CharField(max_length=32, blank=True, null=True)
    decimals = models.BigIntegerField()
    aggregation = models.BigIntegerField()
    connection = models.CharField(max_length=1024, blank=True, null=True)
    comment = models.CharField(max_length=1024, blank=True, null=True)
    rexcore = models.ForeignKey('Rexcore', models.DO_NOTHING, blank=True, null=True)
    rex_value_path = models.CharField(max_length=1024, blank=True, null=True)
    rex_trend_path = models.CharField(max_length=1024, blank=True, null=True)
    range_min = models.FloatField(blank=True, null=True)
    range_max = models.FloatField(blank=True, null=True)
    groups = models.ManyToManyField(
        Grp,
        through='SignalGrp',
        through_fields=('signal', 'grp'),
    )

    class Meta:
        managed = True
        db_table = 'signal'

    def get_all_signals():
        signals = Signal.objects.prefetch_related('groups').all()
        return signals

    def get_one_signal_by_pk(pk):
        return Signal.objects.get(pk=pk)

    def update_all_signals(signals):
        objects = []

        for signal in signals:
            object = Signal(**signal)

            if not object.node_id:
                node = Node.objects.first()
                if node is not None:
                    object.node_id = node.pk

            if not object.tblspace_id:
                space = Tblspace.objects.first()
                if space is not None:
                    object.tblspace_id = space.pk

            if not object.aggregation:
                object.aggregation = 0

            objects.append(object)

        bulk_sync(
            new_models=objects if len(objects) > 0 else Signal.objects.none(),
            filters=[],
            fields=[
                'name',
                'description',
                'type',
                'unit',
                'decimals',
                'rexcore_id',
                'rex_value_path',
            ],
            key_fields=['id']
        )


class SignalGrp(models.Model):
    signal = models.OneToOneField(Signal, models.DO_NOTHING, primary_key=True)
    grp = models.ForeignKey(Grp, models.DO_NOTHING)

    class Meta:
        managed = True
        db_table = 'signal_grp'
        unique_together = (('signal', 'grp'),)


class Tblspace(models.Model):
    id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=32)
    partition = models.CharField(max_length=32)

    class Meta:
        managed = True
        db_table = 'tblspace'
