from rest_framework import status
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.serializers import ValidationError

from django.http import Http404, HttpResponseBadRequest
from django.conf import settings

from api_base.core.aggregator import Aggregator
from api_base.core.bar_preprocessor import BarPreprocessor
from api_base.core.cumulative_preprocessor import CumulativePreprocessor
from api_base.serializers import SignalDataValueSerializer
from api_base.core.exceptions import Http405
from api_base.core.dynamic_import import dynamic_import

from .rex_core_value_reader.common import RexCoreValueReader

import api_base.views as base_views

import abc
import importlib
import json


Node, Signal, Grp, Rexcore = dynamic_import(".models", ["Node", "Signal", "Grp", "Rexcore", ])


class Option(base_views.Option):
    pass


class Version(base_views.Version):
    pass


class SignalList(base_views.SignalList):
    def get_signals(self):
        signals = Signal.get_all_signals()
        return signals

    def put(self, request):
        if hasattr(settings, 'APP_OPTION') \
            and 'signal/edit' in settings.APP_OPTION \
            and settings.APP_OPTION['signal/edit']:
            signals = json.loads(request.body)

            try:
                Signal.update_all_signals(signals)

                return Response(status=status.HTTP_200_OK)
            except Http405:
                return Response(status=status.HTTP_405_METHOD_NOT_ALLOWED)
            except Exception as e:
                return HttpResponseBadRequest(json.dumps({
                    'text': str(e)
                }))
        else:
            return Response(status=status.HTTP_405_METHOD_NOT_ALLOWED)


class SignalDetail(base_views.SignalDetail):
    def get_signal(self, pk):
        try:
            return Signal.get_one_signal_by_pk(pk=pk)
        except Signal.DoesNotExist:
            raise Http404


class GroupList(base_views.GroupList):
    def get_groups(self):
        return Grp.get_all_groups()


class GroupDetail(base_views.GroupDetail):
    def get_group(self, pk):
        try:
            return Grp.get_one_group_by_pk(pk=pk)
        except Grp.DoesNotExist:
            raise Http404


class SignalGroupsList(base_views.SignalGroupsList):
    def get_signal_groups(self, signal_pk):
        return Grp.get_signals_groups(signal_pk)


class RexCoreList(base_views.RexCoreList):
    def get_connections(self):
        return Rexcore.get_all_connections()

    def put(self, request):
        try:
            connections = json.loads(request.body)
            Rexcore.update_all_connections(connections)

            return Response(status=status.HTTP_200_OK)
        except Http405:
            return Response(status=status.HTTP_405_METHOD_NOT_ALLOWED)
        except Exception as e:
            return HttpResponseBadRequest(json.dumps({
                'text': str(e)
            }))


class RexCoreDetail(base_views.RexCoreDetail):
    def get_connection(self, pk):
        return Rexcore.get_one_connection_by_pk(pk)


class BatchSignalLastValue(base_views.BatchSignalLastValue):
    def _create_rexcore_value_reader(self):
        return RexCoreValueReader()


class SignalDataView(base_views.SignalDataView):
    def __init__(self):
        self.signal = None

    def get_object(self, pk):
        try:
            return Signal.get_one_signal_by_pk(pk=pk)
        except Signal.DoesNotExist:
            raise Http404

    def get(self, request, signal_pk):
        signal = self.get_object(signal_pk)
        self.signal = signal

        try:
            context = self.get_context()

            provider_class_name = settings.APP_DATA_PROVIDER # TODO MM validation

            module_name, class_name = provider_class_name.rsplit(".", 1)
            module = importlib.import_module(module_name)
            ProviderClass = getattr(module, class_name)

            provider = ProviderClass()
            result = provider.load(context, signal)

            window_level, values = self.process(context, result)
        except ValidationError as e:
            # TODO MM
            window_level = 0
            values = []

        serializer = SignalDataValueSerializer(values, many=True, read_only=True)
        return Response({
            'id': signal.pk,
            'range_min': signal.range_min,
            'range_max': signal.range_max,
            'values': {
                'level': window_level,
                'data': serializer.data,
            }
        })

    @abc.abstractmethod
    def process(self, context, result):
        pass


class SignalDataAutoView(base_views.SignalDataAutoView, SignalDataView):
    def get_context(self):
        context = super().get_context()

        aggregation = None

        if self.signal.type == 'ANALOG':
            aggregation = self.request.query_params.get('aggregation', None)
            if aggregation is not None and aggregation.isnumeric():
                aggregation = int(aggregation)
                if aggregation > 0:
                    window_segment = aggregation * 60

                    context["window_segment"] = window_segment

        if self.signal.type == 'CUMULATIVE':
            aggregation = self.request.query_params.get('aggregation', None)
            if aggregation is None:
                raise ValidationError('Field aggregation is required!')
            if not aggregation.isnumeric():
                raise ValidationError('Aggregation must be number!')
            aggregation = int(aggregation)
            if aggregation <= 0:
                raise ValidationError('Aggregation must be positive number!')

        context["aggregation"] = aggregation

        return context

    def process(self, context, result):
        if self.signal.type == 'CUMULATIVE':
            preprocessor = CumulativePreprocessor()
            preprocessor.process(context, result)
            return preprocessor.result()
        else:
            aggregator = Aggregator()
            aggregator.agregate(context, result)
            return aggregator.result()


class SignalDataAnalogView(base_views.SignalDataAnalogView, SignalDataView):
    def get_context(self):
        context = super().get_context()

        aggregation = self.request.query_params.get('aggregation', None)
        if aggregation is not None and aggregation.isnumeric():
            aggregation = int(aggregation)
            if aggregation > 0:
                window_segment = aggregation * 60

                context["window_segment"] = window_segment

        return context

    def process(self, context, result):
        aggregator = Aggregator()
        aggregator.agregate(context, result)
        return aggregator.result()


class SignalDataBarView(base_views.SignalDataBarView, SignalDataView):
    def get_context(self):
        context = super().get_context()

        aggregation = self.request.query_params.get('aggregation', None)
        if aggregation is None:
            raise ValidationError('Field aggregation is required!')
        if not aggregation.isnumeric():
            raise ValidationError('Aggregation must be number!')
        aggregation = int(aggregation)
        if aggregation <= 0:
            raise ValidationError('Aggregation must be positive number!')

        context["aggregation"] = aggregation
        context["window_segment"] = aggregation * 60

        return context

    def process(self, context, result):
        preprocessor = BarPreprocessor()
        preprocessor.process(context, result)
        return preprocessor.result()


class SignalDataCumulativeView(base_views.SignalDataCumulativeView, SignalDataView):
    def get_context(self):
        context = super().get_context()

        aggregation = self.request.query_params.get('aggregation', None)
        if aggregation is None:
            raise ValidationError('Field aggregation is required!')
        if not aggregation.isnumeric():
            raise ValidationError('Aggregation must be number!')
        aggregation = int(aggregation)
        if aggregation <= 0:
            raise ValidationError('Aggregation must be positive number!')

        context["aggregation"] = aggregation
        context["window_segment"] = aggregation * 60

        return context

    def process(self, context, result):
        preprocessor = CumulativePreprocessor()
        preprocessor.process(context, result)
        return preprocessor.result()


class SignalDataBoolView(base_views.SignalDataBoolView, SignalDataView):

    def process(self, context, result):
        aggregator = Aggregator()
        aggregator.agregate(context, result)
        return aggregator.result()


class Export(base_views.Export):
    def get_data_provider(self):
        provider_class_name = settings.APP_DATA_PROVIDER # TODO MM validation

        module_name, class_name = provider_class_name.rsplit(".", 1)
        module = importlib.import_module(module_name)
        ProviderClass = getattr(module, class_name)

        return ProviderClass()

    def get_nodes(self):
        nodes = Node.get_all_nodes()
        return nodes

    def get_signals(self):
        signals = Signal.get_all_signals()
        return signals

    def get_signal(self, pk):
        try:
            return Signal.get_one_signal_by_pk(pk=pk)
        except Signal.DoesNotExist:
            raise Http404
