# -*- coding: utf-8 -*-
from django.core.management.base import BaseCommand
from datetime import datetime

from api_base.core.dynamic_import import dynamic_import
from api_database.models import get_data_model

import calendar
import csv


Data, Datatbl, Grp, Node, Signal, SignalGrp, Tblspace \
    = dynamic_import(".models", ["Data", "Datatbl", "Grp", "Node", "Signal", "SignalGrp", "Tblspace"])


class Command(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument(
            '--file',
            type=str,
            required=True,
            help='file'
        )

    def handle(self, *args, **options):
        self.remove_all_values()
        self.generate_from_file(args, options)

    def generate_from_file(self, args, options):
        # node
        node = Node()
        node.pk = 1
        node.name = "NS.LVL.SIGNAL"
        node.save()

        # tbl space
        tblspace = Tblspace()
        tblspace.pk = 1
        tblspace.name = "NS.LVL.SIGNAL_"
        tblspace.save()

        # model
        data_model = None
        signal_value_pk = 0

        # signal map
        signal_map = {}

        # file
        file_path = options["file"]
        with open(file_path) as file:
            reader = csv.reader(file)
            line_count = 0

            for row in reader:
                date = row[0]
                time = row[1]

                stamp_actual = datetime.strptime(
                    date + " " + time,
                    '%Y-%m-%d %H:%M:%S.%f'
                )

                if data_model is None:
                    data_model, signal_value_pk = self.load_model(stamp_actual, tblspace)

                line_count = line_count + 1
                self.print(0, "Line: " + str(line_count))

                signal_group_id = int(row[5])
                if signal_group_id in signal_map:
                    signal = signal_map[signal_group_id]
                else:
                    signal = self.create_signal(
                        signal_group_id,
                        "ANALOG",
                        "SIGNAL " + str(signal_group_id),
                        node,
                        tblspace
                    )
                    signal_map[signal_group_id] = signal

                columns_count = len(row) - 5

                for index in range(1, columns_count):
                    value = row[index + 5]
                    signal_value_pk = signal_value_pk + 1

                    self.create_signal_data(
                        signal,
                        stamp_actual,
                        value,
                        data_model,
                        signal_value_pk

                    )

    def load_model(self, stamp, tblspace):
        # data table
        datatbl_item = None

        ts_year = stamp.strftime("%Y")
        ts_month = stamp.strftime("%m")
        ts_month_range = calendar.monthrange(stamp.year, stamp.month)
        ts_date_from = ts_year + "-" + ts_month + "-" + "01"
        ts_date_to = ts_year + "-" + ts_month + "-" + str(ts_month_range[1]).zfill(2)

        datatbl_item_name = "DATA_TS1_" + ts_year + ts_month

        # data table
        if datatbl_item is None or datatbl_item_name != datatbl_item.name:
            datatbl_item_query = Datatbl.objects \
                .filter(tblspace=tblspace) \
                .filter(date_from=ts_date_from) \
                .filter(date_to=ts_date_to)

            if len(datatbl_item_query) > 0:
                datatbl_item = datatbl_item_query.first()
            else:
                datatbl_item = Datatbl()
                datatbl_item.pk = tblspace.pk  # TODO MM
                datatbl_item.tblspace = tblspace
                datatbl_item.date_from = ts_date_from
                datatbl_item.date_to = ts_date_to
                datatbl_item.name = datatbl_item_name
                datatbl_item.save()

            data_model = get_data_model(Data, datatbl_item.name)

            try:
                latest_value = data_model.objects.latest('id')
                signal_value_pk = latest_value.pk
            except data_model.DoesNotExist:
                signal_value_pk = 0

        return [data_model, signal_value_pk]

    def remove_all_values(self):
        self.print(0, "Deleting old signals ...")

        for datatbl_item in Datatbl.objects.all():
            data_model = get_data_model(Data, datatbl_item.name)
            data_model.objects.all().delete()

        Datatbl.objects.all().delete()
        SignalGrp.objects.all().delete()
        Signal.objects.all().delete()
        Tblspace.objects.all().delete()

        self.print(0, "Done.")

    def create_signal(
        self,
        index,
        signal_type,
        signal_name,
        node,
        tblspace
    ):
        try:
            signal = Signal.objects.get(pk=index)
            return signal
        except Signal.DoesNotExist:
            pass

        # signal
        signal = Signal()
        signal.pk = index
        signal.name = signal_name
        signal.description = signal_name
        signal.node_id = node.pk
        signal.tblspace_id = tblspace.pk
        signal.type = signal_type
        signal.unit = 'unit'
        signal.decimals = 0
        signal.aggregation = 0
        signal.connection = ''
        signal.save()

        self.print(1, "Creating: " + signal.description)

        return signal

    def create_signal_data(
        self,
        signal,
        stamp,
        value,
        data_model,
        signal_value_pk
    ):
        quality = 1.0
        value_final = value

        signal_value = data_model()
        signal_value.pk = signal_value_pk
        signal_value.signal = signal
        signal_value.arcid = 0
        signal_value.dt = stamp
        signal_value.dt_prev = 0
        signal_value.dt_next = 0
        signal_value.value = value_final
        signal_value.value_min = value_final
        signal_value.value_max = value_final
        signal_value.quality = quality
        signal_value.save()

    # ----------------------------------------------------------------------- #
    # HELPERS
    # ----------------------------------------------------------------------- #
    def print(self, indent, msg):
        print("{}{}".format("".rjust(indent * 4, " "), msg))
