# -*- coding: utf-8 -*-
from django.core.management.base import BaseCommand
from datetime import datetime, timedelta
from functools import partial

from api_base.core.dynamic_import import dynamic_import
from api_database.models import get_data_model

import random
import time
import math
import calendar


Data, Datatbl, Grp, Node, Signal, SignalGrp, Tblspace \
    = dynamic_import(".models", ["Data", "Datatbl", "Grp", "Node", "Signal", "SignalGrp", "Tblspace"])


class Command(BaseCommand):

    # ----------------------------------------------------------------------- #
    # CONFIG
    # ----------------------------------------------------------------------- #
    GROUPS = range(1, 10)

    #TIMELINE_DAYS = 90

    #CONF_QUALITY = [(1, 0), (20 * 60, 0.2), (20 * 60, 0.5)]
    #CONF_PERIOD = [5 * 60, 20 * 60, 100 * 60]
    #CONF_STEP = [1 * 60, 5 * 60, 60 * 60]
    #CONF_V_RANGE = [(-200, -150), (-100, +100), (+50, +200)]

    TIMELINE_DAYS = 20

    CONF_QUALITY = [(1, 1), (200 * 5 * 60, 0.8)]
    CONF_PERIOD = [40 * 5 * 60]
    CONF_STEP = [5 * 60]
    CONF_V_RANGE = [(-100, +100, 10)]

    # ----------------------------------------------------------------------- #
    # HANDLE
    # ----------------------------------------------------------------------- #
    def handle(self, *args, **options):
        self.remove_all_values()
        self.create_groups(self.GROUPS)
        self.create_signals()

    # ----------------------------------------------------------------------- #
    # PREPARE
    # ----------------------------------------------------------------------- #
    def remove_all_values(self):
        self.print(0, "Deleting old signals ...")

        for datatbl_item in Datatbl.objects.all():
            data_model = get_data_model(Data, datatbl_item.name)
            data_model.objects.all().delete()

        Datatbl.objects.all().delete()
        SignalGrp.objects.all().delete()
        Signal.objects.all().delete()
        Tblspace.objects.all().delete()

        self.print(0, "Done.")

    # ----------------------------------------------------------------------- #
    # GROUPS
    # ----------------------------------------------------------------------- #
    def create_groups(self, groups_ids):
        self.print(0, "Creating groups ...")

        for index in groups_ids:
            self.create_group(index)

        self.print(0, "Done.")

    def create_group(self, index):
        try:
            grp = Grp.objects.get(pk=index)
            self.print(1, "Group (" + str(index) + ") already exists")
        except Grp.DoesNotExist: # TODO MM
            grp = Grp()
            grp.pk = index
            grp.name = "Group (" + str(index) + ")"
            grp.description = "Very Very Very Looooong Description of Group (" + str(index) + ")"
            grp.save()

            self.print(1, "Creating group (" + str(index) + ")")

    # ----------------------------------------------------------------------- #
    # SIGNALS
    # ----------------------------------------------------------------------- #
    def create_signals(self):
        self.print(0, "Creating signals ...")

        for quality in self.CONF_QUALITY:
            name = {}
            name_quality = "Q:" + str(quality[1])
            name["quality"] = name_quality

            for period in self.CONF_PERIOD:
                name_period = "P" + str(period / 60)
                name["period"] = name_period

                for step in self.CONF_STEP:
                    name_step = "S" + str(step / 60)
                    name["step"] = name_step

                    # ------------------------------------------------------------------- #
                    # BOOL
                    # ------------------------------------------------------------------- #
                    self.create_signal_and_data(
                        "BOOL", "BOOL SIN", name,
                        partial(self.generator_binary_sin, period=period, threshold=0.5),
                        0,
                        partial(self.generator_quality, period=quality[0], threshold=quality[1]),
                        False,
                        self.TIMELINE_DAYS, step
                    )
                    self.create_signal_and_data(
                        "BOOL", "BOOL SAW", name,
                        partial(self.generator_binary_saw, period=period, threshold=0.5),
                        0,
                        partial(self.generator_quality, period=quality[0], threshold=quality[1]),
                        False,
                        self.TIMELINE_DAYS, step
                    )
                    self.create_signal_and_data(
                        "BOOL", "BOOL RND", name,
                        partial(self.generator_binary_random, threshold=0.5),
                        0,
                        partial(self.generator_quality, period=quality[0], threshold=quality[1]),
                        False,
                        self.TIMELINE_DAYS, step
                    )

                    for v_range in self.CONF_V_RANGE:
                        name_range = "(" + str(v_range[0]) + " <-> " + str(v_range[1]) + " / " + str(v_range[2]) + ")"
                        name["range"] = name_range

                        # ------------------------------------------------------------------- #
                        # ANALOG
                        # ------------------------------------------------------------------- #
                        self.create_signal_and_data(
                            "ANALOG", "ANALOG SIN", name,
                            partial(self.generator_analog_sin, period=period, value_min=v_range[0], value_max=v_range[1]),
                            v_range[2],
                            partial(self.generator_quality, period=quality[0], threshold=quality[1]),
                            False,
                            self.TIMELINE_DAYS, step
                        )
                        self.create_signal_and_data(
                            "ANALOG", "ANALOG SAW", name,
                            partial(self.generator_analog_saw, period=period, value_min=v_range[0], value_max=v_range[1]),
                            v_range[2],
                            partial(self.generator_quality, period=quality[0], threshold=quality[1]),
                            False,
                            self.TIMELINE_DAYS, step
                        )
                        self.create_signal_and_data(
                            "ANALOG", "ANALOG RND", name,
                            partial(self.generator_analog_random, value_min=v_range[0], value_max=v_range[1]),
                            v_range[2],
                            partial(self.generator_quality, period=quality[0], threshold=quality[1]),
                            False,
                            self.TIMELINE_DAYS, step
                        )

                        # ------------------------------------------------------------------- #
                        # CUMULATIVE
                        # ------------------------------------------------------------------- #
                        self.create_signal_and_data(
                            "CUMULATIVE", "CUMULATIVE SIN", name,
                            partial(self.generator_analog_sin, period=period, value_min=v_range[0], value_max=v_range[1]),
                            v_range[2],
                            partial(self.generator_quality, period=quality[0], threshold=quality[1]),
                            True,
                            self.TIMELINE_DAYS, step
                        )
                        self.create_signal_and_data(
                            "CUMULATIVE", "CUMULATIVE SAW", name,
                            partial(self.generator_analog_saw, period=period, value_min=v_range[0], value_max=v_range[1]),
                            v_range[2],
                            partial(self.generator_quality, period=quality[0], threshold=quality[1]),
                            True,
                            self.TIMELINE_DAYS, step
                        )
                        self.create_signal_and_data(
                            "CUMULATIVE", "CUMULATIVE RND", name,
                            partial(self.generator_analog_random, value_min=v_range[0], value_max=v_range[1]),
                            v_range[2],
                            partial(self.generator_quality, period=quality[0], threshold=quality[1]),
                            True,
                            self.TIMELINE_DAYS, step
                        )

        self.print(0, "Done.")

    def create_signal_and_data(
        self,
        signal_type,
        signal_name,
        signal_name_add,
        data_generator,
        value_std,
        quality_generator,
        cumulative,
        timeline_days_range,
        step_seconds
    ):
        try:
            latest_signal = Signal.objects.latest('id')
            index = latest_signal.pk
        except Signal.DoesNotExist:
            index = 0

        index = index + 1

        signal = self.create_signal(
            index,
            signal_type,
            signal_name,
            signal_name_add,
            timeline_days_range
        )

        self.create_signal_data(
            signal,
            data_generator,
            value_std,
            quality_generator,
            cumulative,
            timeline_days_range,
            step_seconds
        )

    def create_signal(
        self, 
        index, 
        signal_type, 
        signal_name, 
        signal_name_add, 
        timeline_days_range
    ):
        # stamp options
        stamp_from = datetime.today() - timedelta(days=timeline_days_range)
        stamp_to = datetime.today()

        # node
        node = Node()
        node.pk = index
        node.name = "NS.LVL.SIGNAL_" + str(index)
        node.save()

        # tbl space
        tblspace = Tblspace()
        tblspace.pk = index
        tblspace.name = "NS.LVL.SIGNAL_" + str(index)
        tblspace.save()

        # signal
        signal = Signal()
        signal.pk = index
        signal.name = signal_name + " " \
            + signal_name_add["step"] + "/" \
            + signal_name_add["period"] + " " \
            + signal_name_add["quality"]
        signal.description = "Signal [" \
            + signal_name + " " \
            + signal_name_add["step"] + "/" \
            + signal_name_add["period"] + " " \
            + signal_name_add["quality"] \
            + "] " \
            + ("range: " + signal_name_add["range"] + ", " if "range" in signal_name_add else "") \
            + "ID: " + str(index) + ", " \
            + "timeline: <" + stamp_from.strftime("%Y-%m-%d, %H:%M:%S") + ", " + stamp_to.strftime("%Y-%m-%d, %H:%M:%S") + ">, "
        signal.node_id = node.pk
        signal.tblspace_id = tblspace.pk
        signal.type = signal_type
        signal.unit = 'unit'
        signal.decimals = 0
        signal.aggregation = 0
        signal.connection = ''
        signal.save()

        self.print(1, "Creating: " + signal.description)

        self.print(2, "Creating signals groups ...")

        for group in Grp.objects.all():
            if random.choice([True, False, False]):
                signal_grp = SignalGrp()
                signal_grp.signal = signal
                signal_grp.grp = group
                signal_grp.save()

                self.print(3, "Added group (" + str(group.pk) + ") into signal (" + str(index) + ")")

        return signal

    def create_signal_data(
        self,
        signal,
        data_generator,
        value_std,
        quality_generator,
        cumulative,
        timeline_days_range,
        step_seconds
    ):
        self.print(2, "Creating signal (" + str(signal.pk) + ") values ...")

        # data table
        datatbl_item = None

        # stamp options
        stamp_from = datetime.today() - timedelta(days=timeline_days_range)
        stamp_to = datetime.today()
        stamp = stamp_from

        # last value
        last_value = None

        while stamp <= stamp_to:
            stamp = stamp + timedelta(seconds=step_seconds)

            ts_year = stamp.strftime("%Y")
            ts_month = stamp.strftime("%m")
            ts_month_range = calendar.monthrange(stamp.year, stamp.month)
            ts_date_from = ts_year + "-" + ts_month + "-" + "01"
            ts_date_to = ts_year + "-" + ts_month + "-" + str(ts_month_range[1]).zfill(2)

            datatbl_item_name = "DATA_TS1_" + ts_year + ts_month

            # data table
            if datatbl_item is None or datatbl_item_name != datatbl_item.name:
                datatbl_item_query = Datatbl.objects \
                    .filter(tblspace=signal.tblspace) \
                    .filter(date_from=ts_date_from) \
                    .filter(date_to=ts_date_to)

                if len(datatbl_item_query) > 0:
                    datatbl_item = datatbl_item_query.first()
                else:
                    datatbl_item = Datatbl()
                    datatbl_item.pk = signal.pk
                    datatbl_item.tblspace = signal.tblspace
                    datatbl_item.date_from = ts_date_from
                    datatbl_item.date_to = ts_date_to
                    datatbl_item.name = datatbl_item_name
                    datatbl_item.save()

                data_model = get_data_model(Data, datatbl_item.name)
            
                try:
                    latest_value = data_model.objects.latest('id')
                    signal_value_pk = latest_value.pk
                except data_model.DoesNotExist:
                    signal_value_pk = 0

            signal_value_pk = signal_value_pk + 1

            value = data_generator(stamp)
            quality = quality_generator(stamp)

            if cumulative:
                if last_value is None:
                    value_final = max(0, value)
                    last_value = value_final
                else:
                    value_final = last_value + max(0, value)
                    last_value = value_final
            else:
                value_final = value

            signal_value = data_model()
            signal_value.pk = signal_value_pk
            signal_value.signal = signal
            signal_value.arcid = 0
            signal_value.dt = stamp
            signal_value.dt_prev = 0
            signal_value.dt_next = 0
            signal_value.value = value_final
            signal_value.value_min = value_final + random.randint(-value_std, +value_std)
            signal_value.value_max = value_final + random.randint(-value_std, +value_std)
            signal_value.quality = quality
            signal_value.save()

    # ----------------------------------------------------------------------- #
    # GENERATORS
    # ----------------------------------------------------------------------- #
    def generator_analog_sin(self, dt, period, value_min, value_max):
        frequency = 2 * math.pi * dt.timestamp() / period

        value_range = value_max - value_min
        value_center = value_min + value_range / 2
        value = value_center + value_range / 2 * math.sin(frequency)

        return value

    def generator_analog_saw(self, dt, period, value_min, value_max):
        value_range = value_max - value_min
        value_coef = value_range / period
        value = (dt.timestamp() % period) * value_coef + value_min

        return value

    def generator_analog_random(self, dt, value_min, value_max):
        value = random.randint(
            value_min,
            value_max
        )

        return value

    def generator_binary_sin(self, dt, period, threshold):
        value = self.generator_analog_sin(dt, period, 0, 1)
        return 1 if value >= threshold else 0

    def generator_binary_saw(self, dt, period, threshold):
        value = self.generator_analog_saw(dt, period, 0, 1)
        return 1 if value >= threshold else 0

    def generator_binary_random(self, dt, threshold):
        value = self.generator_analog_random(dt, 0, 1)
        return 1 if value >= threshold else 0

    def generator_quality(self, dt, period, threshold):
        value = self.generator_binary_sin(dt, period, 1 - threshold)
        return value

    # ----------------------------------------------------------------------- #
    # HELPERS
    # ----------------------------------------------------------------------- #
    def print(self, indent, msg):
        print("{}{}".format("".rjust(indent * 4, " "), msg))
