from ctypes import addressof, c_uint8, memmove, sizeof
from os import path
from typing import List

from ..core.buffer import Buffer
from ..core.read_state import ReadState
from ..core.structs import ACI_WITH_DATE, AR_COMMON_ITEM
from ..core.helpers import ARCODE, ARCLASS, XFAILED, XBYTEARR2XWORD, XBYTEARR2XDWORD, valid_item_code, get_alarm_size, get_group_size
from ..core.constants import ALL_INTEGRITY_ERROR, ARC_SYSTEM_ALARM, ARS_ARC_EVENTS, ATF_DATE_MARK_FLAG, XR_AR_INTEGRITY_ERROR, \
    XR_OK, MINIMAL_REM_COUNT, BUFFER_SIZE, XR_AR_NO_MORE_ITEMS, SOF_TSTAMP, ARS_DATE_MARK, SOF_CODE, SOF_LEVEL, SOF_ALARMID, \
    ARC_ALARM_LAST, ARC_ALARM_ACKN, ARC_STRING_ALARM, SIZEOF_XWORD, SIZEOF_XDWORD, SOF_TCNI, SIZEOF_SYSTEM_ALARM, SOF_TCLI
from ..writer.writer import BaseWriter

import logging


class Parser:
    def __init__(self):
        self.buffer = Buffer()
        self.read_state = ReadState()
        self.fstream = None

    def export_files(self, writer: BaseWriter, files: List[str]):
        for file in files:
            local_result = XR_OK
            acid = ACI_WITH_DATE()
            can_continue = True

            if not self._open_file(file):
                continue

            self._load()

            while local_result >= XR_OK:
                local_result = self._read_next_item_from_data(self.buffer, self.read_state, acid.aci)

                if local_result == XR_OK:
                    acid.date = self.read_state.get_rs_date()
                    acid.marked = False

                    can_continue = writer.export_value(acid)
                    if not can_continue:
                        break

                    if self.buffer.len() < MINIMAL_REM_COUNT:
                        self._load()

                elif XFAILED(local_result):
                    if local_result == XR_AR_INTEGRITY_ERROR:
                        acid.date = self.read_state.get_rs_date()
                        acid.marked = False
                        acid.aci.code = ARCODE(ARC_SYSTEM_ALARM, ARS_ARC_EVENTS)
                        acid.aci.lev_cnt = ALL_INTEGRITY_ERROR
                        writer.export_value(acid)

                    raise Exception("Error occured while reading file '" + file + "'.")

            self._close_current_file()

            if not can_continue:
                break

    def export_files_as_generator(self, writer: BaseWriter, files: List[str]):
        for file in files:
            local_result = XR_OK
            acid = ACI_WITH_DATE()

            if not self._open_file(file):
                continue

            self._load()

            while local_result >= XR_OK:
                local_result = self._read_next_item_from_data(self.buffer, self.read_state, acid.aci)

                if local_result == XR_OK:
                    acid.date = self.read_state.get_rs_date()
                    acid.marked = False
                    yield from writer.export_value(acid)

                    if self.buffer.len() < MINIMAL_REM_COUNT:
                        self._load()

                elif XFAILED(local_result):
                    if local_result == XR_AR_INTEGRITY_ERROR:
                        acid.date = self.read_state.get_rs_date()
                        acid.marked = False
                        acid.aci.code = ARCODE(ARC_SYSTEM_ALARM, ARS_ARC_EVENTS)
                        acid.aci.lev_cnt = ALL_INTEGRITY_ERROR
                        yield from writer.export_value(acid)

                    raise Exception("Error occured while reading file '" + file + "'.")

            self._close_current_file()

    def _open_file(self, file):
        self._close_current_file()

        if not path.exists(file):
            raise Exception("File does not exist!")

        self.fstream = open(file, "rb")

        self.read_state.set_rs_date(0)

        return True

    def _close_current_file(self):
        if self.fstream and not self.fstream.closed:
            self.fstream.close()
            self.fstream = None

    def _load(self):
        if self.fstream and not self.fstream.closed:
            chunkSize = BUFFER_SIZE - self.buffer.len()
            bytes_read = self.fstream.read(chunkSize)  # TODO MM
            if bytes_read:
                bytes = bytearray(bytes_read)
                self.buffer.extend(bytes)
            else:
                self.fstream.close()
                self.fstream = None

    def _read_next_item_from_data(self, buffer: Buffer, ARS: ReadState, ACI: AR_COMMON_ITEM):
        # XASSERT(buffer != NULL)
        # XASSERT(bufferSize != NULL)
        # XASSERT(*bufferSize > 0)
        # XASSERT(ARS != NULL)
        # XASSERT(ACI != NULL)

        return self._data_get_item(buffer, ARS, ACI)

    def _data_get_item(self, buffer: Buffer, ARS: ReadState, ACI: AR_COMMON_ITEM):
        # Sets *ACI to the archive item specified by *ARS in data array *buffer.
        # Calls DataGetAnyItem() and skips date mark. Returned number of bytes
        # includes potentially skipped date mark.
        nSize = 0

        # XASSERT(buffer != NULL)
        # XASSERT(bufferSize != NULL)
        # XASSERT(*bufferSize > 0)
        # XASSERT(ARS != NULL)
        # XASSERT(ACI != NULL)

        while True:
            nRes = self._data_get_any_item(buffer, ACI)
            if nRes < 0:
                return nRes

            nSize += nRes

            if (ACI.ticks_hi & ATF_DATE_MARK_FLAG) == 0:
                break

            # XASSERT(ACI.code == ARCODE(ARC_SYSTEM_ALARM, ARS_DATE_MARK) \
            # and ACI.ticks_hi == ATF_DATE_MARK_FLAG and ACI.ticks_lo <= 0xFFFF \
            # and ACI.ticks_lo >= ARS.get_rs_date())

            ARS.set_rs_date(ACI.ticks_lo)

        return XR_OK

    def _data_get_any_item(self, buffer: Buffer, ACI: AR_COMMON_ITEM):
        # Sets *ACI to the archive item pointed by *ppDtPos to data array *buffer.
        # Returns negative error code or positive number of bytes read from archive.
        # This function returns also date mark.

        if buffer.len() == 0:
            return XR_AR_NO_MORE_ITEMS  # we reach the data end

        if buffer.byte_at(0) == 0xFF:
            return XR_AR_NO_MORE_ITEMS  # end of data before end of array

        ticks = self._parse_ticks(buffer, ACI)
        if ticks is not None:
            return ticks

        code = self._parse_code(buffer, ACI)
        if code < 0:
            return code

        if code != ARC_SYSTEM_ALARM:
            ACI.ID = XBYTEARR2XWORD(buffer)
            buffer.delete(SOF_ALARMID)

            if ((code <= ARC_ALARM_LAST) or (code == ARC_ALARM_ACKN)):
                return self._parse_alarm(buffer, ACI, code)
            else:
                return self._parse_group(buffer, ACI, code)

        else:
            return SIZEOF_SYSTEM_ALARM

    def _parse_ticks(self, buffer: Buffer, ACI: AR_COMMON_ITEM):
        ACI.ticks_hi = XBYTEARR2XWORD(buffer)
        buffer.delete(SIZEOF_XWORD)

        ACI.ticks_lo = XBYTEARR2XDWORD(buffer)
        buffer.delete(SIZEOF_XDWORD)

        if (ACI.ticks_hi & ATF_DATE_MARK_FLAG):
            # date mark
            if (ACI.ticks_hi != ATF_DATE_MARK_FLAG or ACI.ticks_lo > 0xFFFF):
                return XR_AR_INTEGRITY_ERROR

            ACI.code = ARCODE(ARC_SYSTEM_ALARM, ARS_DATE_MARK)
            ACI.lev_cnt = 0

            return SOF_TSTAMP

        return None

    def _parse_code(self, buffer: Buffer, ACI: AR_COMMON_ITEM):
        ACI.code = buffer.byte_at(0)
        buffer.delete(SOF_CODE)

        code = ARCLASS(ACI.code)
        if not valid_item_code(code):
            return XR_AR_INTEGRITY_ERROR

        ACI.lev_cnt = buffer.byte_at(0)
        buffer.delete(SOF_LEVEL)

        return code

    def _parse_alarm(self, buffer: Buffer, ACI: AR_COMMON_ITEM, code):
        #  Get an alarm
        nSize = get_alarm_size(code)

        if code != ARC_STRING_ALARM:
            count = nSize - SOF_TCLI
            if count > 0:
                tmp = buffer.sub_reverse(count)
                local_buffer = (c_uint8 * count).from_buffer(tmp)
                memmove(addressof(ACI.au), local_buffer, count)
                del local_buffer
                buffer.delete(count)

        else:
            wLen = XBYTEARR2XWORD(buffer)
            buffer.delete(SIZEOF_XWORD)

            ACI.length = wLen+1
            #  TODO MM
            # ACI.au.alu.xString = allocstr(ACI.length)
            # if (ACI.au.alu.xString == NULL)
            #     return XR_NOT_ENOUGH_MEMORY

            # // Read XSTRING itself
            # memcpy(ACI.au.alu.xString, *ppDtPos, wLen)
            # ACI.au.alu.xString[wLen] = _XT('\0')

            buffer.delete(wLen)
            nSize += wLen + SIZEOF_XWORD

        return nSize

    def _parse_group(self, buffer: Buffer, ACI: AR_COMMON_ITEM, code):
        #  Get a group
        nSize = get_group_size(code, ACI.lev_cnt)
        if (nSize > sizeof(ACI.au.agu) + SOF_TCNI):
            return XR_AR_INTEGRITY_ERROR

        count = nSize - SOF_TCNI
        count = int(count)  # TODO MM
        if count > 0:
            tmp = buffer.sub_reverse(count)
            local_buffer = (c_uint8 * count).from_buffer(tmp)
            memmove(addressof(ACI.au.agu), local_buffer, count)
            del local_buffer
            buffer.delete(count)

        return nSize
