from api_impl_rex_archives.arc_reader.writer.signal_table_creator \
    import SignalTableCreator
from rest_framework import viewsets, status
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.serializers import ValidationError

from django.http import Http404
from django.conf import settings

from api_base.core.exceptions import Http405
from api_base.core.aggregator import Aggregator
from api_base.core.bar_preprocessor import BarPreprocessor
from api_base.core.cumulative_preprocessor import CumulativePreprocessor
from api_base.serializers import SignalDataValueSerializer
from api_base.authentication import IsAuthenticated
import api_base.views as base_views

from .signal_explorer import SignalExplorer, SignalExplorerWriter
from .data_provider.from_archives import DataProvider

import datetime
import json
import logging
import abc


class Option(base_views.Option):
    pass


class Version(base_views.Version):
    pass


class SignalList(base_views.SignalList):
    def get_signals(self):
        explorer = SignalExplorer(settings.ARCHIVES_SIGNAL_META_FILE_PATH)
        return explorer.get_signals()

    def put(self, request):
        signals = json.loads(request.body)

        explorer = SignalExplorerWriter(settings.ARCHIVES_SIGNAL_META_FILE_PATH)
        explorer.update(signals)

        return Response(status=status.HTTP_200_OK)


class BatchSignalCreateMetadataTable(APIView):
    permission_classes = [IsAuthenticated]

    def get(self, request):
        provider = SignalTableCreator(settings.ARCHIVES_ROOT_DIR)
        signals = provider.load()

        explorer = SignalExplorerWriter(settings.ARCHIVES_SIGNAL_META_FILE_PATH)
        explorer.update(signals)

        return Response(status=status.HTTP_200_OK)


class SignalDetail(base_views.SignalDetail):
    def get_signal(self, pk):
        explorer = SignalExplorer(settings.ARCHIVES_SIGNAL_META_FILE_PATH)
        signal = explorer.get_signal(pk)
        if signal:
            return signal
        else:
            raise Http404


class GroupList(base_views.GroupList):
    def get_groups(self):
        return []


class GroupDetail(base_views.GroupDetail):
    def get_group(self, pk):
        raise Http404


class RexCoreList(base_views.RexCoreList):
    def get_connections(self):
        return []


class RexCoreDetail(base_views.RexCoreDetail):
    def get_connection(self, pk):
        raise Http404


class BatchSignalLastValue(base_views.BatchSignalLastValue):
    def get(self, request):
        raise Http405


class SignalGroupsList(base_views.SignalGroupsList):
    def get_signal_groups(self, signal_pk):
        return []


class SignalDataView(base_views.SignalDataView):
    def __init__(self):
        self.signal = None

    def get_object(self, pk):
        explorer = SignalExplorer(settings.ARCHIVES_SIGNAL_META_FILE_PATH)
        signal = explorer.get_signal(pk)
        if signal:
            return signal
        else:
            raise Http404

    def get_context(self):
        dt_start = self.request.query_params.get('dt_start', None)
        if dt_start is None:
            raise ValidationError('Field dt_start is required!')

        dt_start = datetime.datetime.strptime(dt_start, '%Y-%m-%d-%H-%M-%S') # TODO MM format
        if dt_start is None:
            raise ValidationError('dt_start has incorect format!')

        dt_end = self.request.query_params.get('dt_end', None)
        if dt_end is None:
            raise ValidationError('Field dt_end is required!')

        dt_end = datetime.datetime.strptime(dt_end, '%Y-%m-%d-%H-%M-%S') # TODO MM format
        if dt_end is None:
            raise ValidationError('dt_end has incorect format!')

        dt_diff = dt_end - dt_start
        dt_diff_seconds = dt_diff.total_seconds() # TODO MM move into helper function
        if dt_diff_seconds <= 0:
            raise ValidationError('Date range is not valid!')

        level = self.request.query_params.get('level', None)
        if level is None:
            raise ValidationError('Field level is required!')
        if not level.isnumeric():
            raise ValidationError('Level must be number!')
        level = int(level)
        if level <= 0:
            raise ValidationError('Level must be positive number!')

        window_segment = int(dt_diff_seconds / level) + 1 # TODO MM move into helper function
        if window_segment <= 0:
            raise ValidationError('Date range is not valid!')

        aggregation = None
        if self.signal.type == 'CUMULATIVE': # TODO MM bug
            aggregation = self.request.query_params.get('aggregation', None)
            if aggregation is None:
                raise ValidationError('Field aggregation is required!')
            if not aggregation.isnumeric():
                raise ValidationError('Aggregation must be number!')
            aggregation = int(aggregation)
            if aggregation <= 0:
                raise ValidationError('Aggregation must be positive number!')

        return {
            'dt_start': dt_start,
            'dt_end': dt_end,
            'aggregation': aggregation,
            'level': level,
            'window_segment': window_segment,
        }

    def get(self, request, signal_pk):
        signal = self.get_object(signal_pk)
        self.signal = signal

        try:
            context = self.get_context()

            provider = DataProvider()
            result = provider.load(context, signal)

            window_level, values = self.process(context, result)
        except ValidationError as e:
            raise e
            # TODO MM
            window_level = 0
            values = []

        serializer = SignalDataValueSerializer(values, many=True, read_only=True)
        return Response({
            'id': signal.id,
            'values': {
                'level': window_level,
                'data': serializer.data,
            }
        })

    @abc.abstractmethod
    def process(self, context, result):
        pass


class SignalDataAutoView(base_views.SignalDataAutoView, SignalDataView):
    def get_context(self):
        context = super().get_context()

        aggregation = None

        if self.signal.type == 'ANALOG':
            aggregation = self.request.query_params.get('aggregation', None)
            if aggregation is not None and aggregation.isnumeric():
                aggregation = int(aggregation)
                if aggregation > 0:
                    window_segment = aggregation * 60

                    context["window_segment"] = window_segment

        if self.signal.type == 'CUMULATIVE':
            aggregation = self.request.query_params.get('aggregation', None)
            if aggregation is None:
                raise ValidationError('Field aggregation is required!')
            if not aggregation.isnumeric():
                raise ValidationError('Aggregation must be number!')
            aggregation = int(aggregation)
            if aggregation <= 0:
                raise ValidationError('Aggregation must be positive number!')

        context["aggregation"] = aggregation

        return context

    def process(self, context, result):
        if self.signal.type == 'CUMULATIVE':
            preprocessor = CumulativePreprocessor()
            preprocessor.process(context, result)
            return preprocessor.result()
        else:
            aggregator = Aggregator()
            aggregator.agregate(context, result)
            return aggregator.result()


class SignalDataAnalogView(base_views.SignalDataAnalogView, SignalDataView):
    def get_context(self):
        context = super().get_context()

        aggregation = self.request.query_params.get('aggregation', None)
        if aggregation is not None and aggregation.isnumeric():
            aggregation = int(aggregation)
            if aggregation > 0:
                window_segment = aggregation * 60

                context["window_segment"] = window_segment

        return context

    def process(self, context, result):
        aggregator = Aggregator()
        aggregator.agregate(context, result)
        return aggregator.result()


class SignalDataBarView(base_views.SignalDataBarView, SignalDataView):
    def get_context(self):
        context = super().get_context()

        aggregation = self.request.query_params.get('aggregation', None)
        if aggregation is None:
            raise ValidationError('Field aggregation is required!')
        if not aggregation.isnumeric():
            raise ValidationError('Aggregation must be number!')
        aggregation = int(aggregation)
        if aggregation <= 0:
            raise ValidationError('Aggregation must be positive number!')

        context["aggregation"] = aggregation
        context["window_segment"] = aggregation * 60

        return context

    def process(self, context, result):
        preprocessor = BarPreprocessor()
        preprocessor.process(context, result)
        return preprocessor.result()


class SignalDataCumulativeView(base_views.SignalDataCumulativeView, SignalDataView):
    def get_context(self):
        context = super().get_context()

        aggregation = self.request.query_params.get('aggregation', None)
        if aggregation is None:
            raise ValidationError('Field aggregation is required!')
        if not aggregation.isnumeric():
            raise ValidationError('Aggregation must be number!')
        aggregation = int(aggregation)
        if aggregation <= 0:
            raise ValidationError('Aggregation must be positive number!')

        context["aggregation"] = aggregation
        context["window_segment"] = aggregation * 60

        return context

    def process(self, context, result):
        preprocessor = CumulativePreprocessor()
        preprocessor.process(context, result)
        return preprocessor.result()


class SignalDataBoolView(base_views.SignalDataBoolView, SignalDataView):

    def process(self, context, result):
        aggregator = Aggregator()
        aggregator.agregate(context, result)
        return aggregator.result()


class Export(base_views.Export):
    def get_data_provider(self):
        return DataProvider()

    def get_nodes(self):
        return []

    def get_signals(self):
        explorer = SignalExplorer(settings.ARCHIVES_SIGNAL_META_FILE_PATH)
        return explorer.get_signals()

    def get_signal(self, pk):
        explorer = SignalExplorer(settings.ARCHIVES_SIGNAL_META_FILE_PATH)
        signal = explorer.get_signal(pk)
        if signal:
            return signal
        else:
            raise Http404
